--*******************************************************************************************************************************
--
-- Description: This module rate buffers data for a DMA write transaction from the channel producing data to the EZDMA2 core.
--					
-- Rev History: 09-04-2013:	Inital release.
--
--*******************************************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.all;

use work.model_specific_package.all;
use work.transceiver_package.all;

entity dma_write_fifo is
port (
    -- Reset and clocks.
	clear 				: in std_logic;										-- Active high synchronous FIFO clear.
	clk 				: in std_logic;										-- System clock supplied by the bus interface.
	rx_clk 				: in std_logic;										-- Receiver clock coincident with the input sample data.
	
    -- Input data bus and associated command/status signals.
	data_in				: in std_logic_vector(63 downto 0);					-- Input bus carrying data to perform a DMA write transaction.
	enable_64bit		: in std_logic_vector((AXI_WIDTH/64)-1 downto 0);	-- Enables to write data into the FIFO.
	fifo_error			: out std_logic;									-- Logic one indicates a FIFO overflow condition.
	
    -- DMA data bus and associated command/status signals.                                                                                      
	dma_enable			: in std_logic;										-- Logic one enables DMA transfers.
	dma_fifo_rd			: in std_logic;										-- Logic one indicates that the PLDA_EZDMA2 core is consuming data from a DMA write request.
	dma_rddata			: out std_logic_vector(AXI_WIDTH-1 downto 0);		-- DMA local data bus used for write transactions.
	dma_burst_size		: in std_logic_vector(13 downto 0);					-- Number of bytes to transfer in each DMA burst.
	dma_go				: out std_logic										-- Initiate a DMA write transaction when enough data is available to satisfy a burst.
);
end dma_write_fifo;

architecture structural of dma_write_fifo is

	signal ovf_error	: std_logic_vector((AXI_WIDTH/64)-1 downto 0);		-- FIFO overflow flags.
	signal data_count	: std_logic_vector((11*AXI_WIDTH/64)-1 downto 0);	-- Depth of data currently in the FIFO.
	signal rd_en_reg	: std_logic;										-- Registered read enable from the EZDMA2 core.
	signal bridge_empty	: std_logic_vector((AXI_WIDTH/64)-1 downto 0);		-- Logic one indicates that the FIFO bridge has no data to transfer.
	signal bridge_full	: std_logic_vector((AXI_WIDTH/64)-1 downto 0);		-- Logic one indicates that the FIFO bridge cannot accept another transfer.
	signal bridge_data	: std_logic_vector(AXI_WIDTH-1 downto 0);			-- Data crossing the FIFO bridge.
	signal bridge_en	: std_logic;										-- Logic one initiates a bridge transfer.	
	
begin

--**************************************************** FIFO INSTANTIATIONS *****************************************************

    -- Parallel FIFOs (8 kbytes each) are used to interface with the channel producing DMA data.
	RX_FIFO_X2 : for i in 0 to (AXI_WIDTH/64)-1 generate
		RX_FIFO : entity work.fifo_channel_1kx64
		port map (
			rst			=> clear,
			wr_clk		=> rx_clk,
			rd_clk		=> clk,
			din			=> data_in,
			wr_en		=> enable_64bit(i),  
			rd_en		=> bridge_en,
			dout		=> bridge_data((64*i)+63 downto 64*i),
			full		=> open,
			overflow	=> ovf_error(i),
			empty		=> bridge_empty(i),
			valid		=> open,
			underflow	=> open
		);
	end generate;
	
	-- Transfer data between FIFOs when possible.
	bridge_en <= '1' when to_integer(unsigned(bridge_empty)) = 0 and to_integer(unsigned(bridge_full)) = 0 else '0';
	
    -- Parallel FIFOs (8 kbytes each) are used to interface with the PLDA EZDMA2 core.
	DMA_WR_X2 : for i in 0 to (AXI_WIDTH/64)-1 generate
		WR_FIFO : entity work.fifo_dma_1kx64
		port map (
			clk				=> clk,
			rst				=> clear,
			din				=> bridge_data((64*i)+63 downto 64*i),
			wr_en			=> bridge_en,  
			rd_en			=> dma_fifo_rd,
			dout			=> dma_rddata((64*i)+63 downto 64*i),
			full			=> bridge_full(i),
			empty			=> open,
			data_count		=> data_count((11*i)+10 downto 11*i)
		);
	end generate;

--***************************************************** PORT ASSIGNMENTS *******************************************************

	-- Initiate a DMA transfer when there is enough data available to satisfy a burst (64-bit AXI).
	RX_GO_X1 : if AXI_WIDTH/64 = 1 generate
		dma_go <= dma_enable when (data_count(10 downto 0) >= dma_burst_size(13 downto 3)) else '0';
	end generate;
	
	-- Initiate a DMA transfer when there is enough data available to satisfy a burst (128-bit AXI).
	RX_GO_X2 : if AXI_WIDTH/64 = 2 generate
		dma_go <= dma_enable when (data_count(10 downto 0) >= '0' & dma_burst_size(13 downto 4)) and (data_count(21 downto 11) >= '0' & dma_burst_size(13 downto 4)) else '0';
	end generate;
	
	-- Report an overflow condition in any FIFO.
	fifo_error <= '1' when to_integer(unsigned(ovf_error)) > 0 else '0';
	
end structural;
